# Copyright (c) 2012-2013, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of Intel Corporation nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""ACPI module."""

import bits
import _bits
import struct
import bitfields
from cpudetect import cpulib
from collections import namedtuple, OrderedDict
import itertools
import textwrap
import ttypager

_wrapper = textwrap.TextWrapper(width=78, subsequent_indent='  ')
_wrapper_indentall = textwrap.TextWrapper(width=78, initial_indent='  ', subsequent_indent='  ')

def print_wrap(str):
    print _wrap(str)

def _wrap(str):
    def __wrap():
        wrapper = _wrapper
        for line in str.split("\n"):
            for wrapped_line in wrapper.wrap(line):
                yield wrapped_line
            wrapper = _wrapper_indentall
    return '\n'.join(__wrap())

from _bits import _acpi_get_table_by_index as get_table_by_index
from _bits import _acpi_install_interface as install_interface
from _bits import _acpi_objpaths as get_objpaths
from _bits import _acpi_remove_interface as remove_interface
from _bits import _acpi_terminate as terminate

def _id(v):
    return v

class TableParseException(Exception): pass

class AcpiBuffer(str):
    def __repr__(self):
        return "AcpiBuffer(" + ' '.join("{:02x}".format(ord(c)) for c in self) + ")"

    def __str__(self):
        return repr(self)

class AcpiPower(namedtuple('acpi_power', ['SystemLevel', 'ResourceOrder'])):
    __slots__ = ()

    def __repr__(self):
        return "acpi_power(SystemLevel={SystemLevel:#x}, ResourceOrder={ResourceOrder:#x})".format(**self._asdict())

class AcpiProcessor(namedtuple('acpi_processor', ['ProcId', 'PblkAddress', 'PblkLength'])):
    __slots__ = ()

    def __repr__(self):
        return "acpi_processor(ProcId={ProcId:#x}, PblkAddress={PblkAddress:#x}, PblkLength={PblkLength})".format(**self._asdict())

ACPI_TYPE_INTEGER = 0x01
ACPI_TYPE_STRING = 0x02
ACPI_TYPE_BUFFER = 0x03
ACPI_TYPE_PACKAGE = 0x04
ACPI_TYPE_POWER = 0x0B
ACPI_TYPE_PROCESSOR = 0x0C

_acpi_object_types = {
    ACPI_TYPE_INTEGER: _id,
    ACPI_TYPE_STRING: _id,
    ACPI_TYPE_BUFFER: AcpiBuffer,
    ACPI_TYPE_PACKAGE: (lambda t: tuple(_acpi_object_to_python(v) for v in t)),
    ACPI_TYPE_POWER: (lambda args: AcpiPower(*args)),
    ACPI_TYPE_PROCESSOR: (lambda args: AcpiProcessor(*args)),
}

def _acpi_object_to_python(acpi_object):
    if acpi_object is None:
        return None
    object_type, value = acpi_object
    return _acpi_object_types[object_type](value)

def _acpi_object_from_python(obj):
    if isinstance(obj, (int, long)):
        return (ACPI_TYPE_INTEGER, obj)
    # Must check AcpiBuffer before str, since AcpiBuffer derives from str
    if isinstance(obj, AcpiBuffer):
        return (ACPI_TYPE_BUFFER, obj)
    if isinstance(obj, str):
        return (ACPI_TYPE_STRING, obj)
    if isinstance(obj, AcpiPower):
        return (ACPI_TYPE_POWER, obj)
    if isinstance(obj, AcpiProcessor):
        return (ACPI_TYPE_PROCESSOR, obj)
    # Must check tuple after any namedtuples, since namedtuples derive from tuple
    if isinstance(obj, tuple):
        return (ACPI_TYPE_PACKAGE, obj)

def evaluate(pathname, *args):
    """Evaluate an ACPI method and return the result."""
    return _acpi_object_to_python(_bits._acpi_eval(pathname, tuple(_acpi_object_from_python(arg) for arg in args)))

def scope(path):
    try:
        prefix, _ = path.rsplit('.', 1)
        return prefix
    except ValueError:
        return "/"

class UnpackError(Exception):
    pass

class Unpackable(object):
    def __init__(self, data, offset=0, size=None):
        self.data = data
        data_size = len(data)
        if offset > data_size:
            raise UnpackError("Unpackable.__init__: offset={} but len(data)={}".format(offset, data_size))
        self.offset = offset
        if size is None:
            self.size = data_size
        else:
            self.size = offset + size
            if self.size > data_size:
                raise UnpackError("Unpackable.__init__: offset+size={} but len(data)={}".format(self.size, data_size))

    def _check_unpack(self, size):
        if self.offset + size > self.size:
            raise UnpackError("Unpackable: Attempted to unpack {} bytes, but only {} bytes remaining".format(size, self.size - self.offset))

    def skip(self, size):
        self._check_unpack(size)
        self.offset += size

    def unpack(self, fmt):
        try:
            l = struct.calcsize(fmt)
            self._check_unpack(l)
            value = struct.unpack_from(fmt, self.data, self.offset)
            self.offset += l
            return value
        except struct.error as e:
            raise UnpackError("Unpackable.unpack: " + str(e))

    def unpack_one(self, fmt):
        return self.unpack(fmt)[0]

    def unpack_peek(self, fmt):
        try:
            l = struct.calcsize(fmt)
            self._check_unpack(l)
            return struct.unpack_from(fmt, self.data, self.offset)
        except struct.error as e:
            raise UnpackError("Unpackable.unpack_peek: " + str(e))

    def unpack_peek_one(self, fmt):
        return self.unpack_peek(fmt)[0]

    def unpack_raw(self, size):
        """Unpack the specified number of bytes as a str"""
        self._check_unpack(size)
        self.offset += size
        return self.data[self.offset:self.offset+size]

    def unpack_rest(self):
        """Return the remainder of the unpackable as a str"""
        offset = self.offset
        self.offset = self.size
        return self.data[offset:self.size]

    def unpack_unpackable(self, size):
        """Unpack the specified number of bytes as an Unpackable"""
        u = Unpackable(self.data, self.offset, size)
        self.offset += size
        return u

    def at_end(self):
        return self.offset == self.size

def parse_table(signature, instance=1):
    data = get_table(signature, instance)
    if data is None:
        return None
    u = Unpackable(data)
    return globals()[signature].unpack(u)

def make_compat_parser(signature):
    def parse(printflag=False, instance=1):
        table = parse_table(signature, instance)
        if table is None:
            return None
        if printflag:
            ttypager.ttypager(_wrap(repr(table)))
        return table
    return parse

class StructError(Exception):
    pass

class Struct(object):
    def __new__(cls, i):
        me = super(Struct, cls).__new__(cls)
        me.fields = OrderedDict()
        for field in i:
            if len(field) == 2:
                key, value = field
                if isinstance(value, (int, long)) and not isinstance(value, bool):
                    fmt = "{:#x}".format
                else:
                    fmt = "{!r}".format
            elif len(field) == 3:
                key, value, fmt = field
                if isinstance(fmt, str):
                    fmt = fmt.format
                elif not callable(fmt):
                    raise StructError("Internal error: Expected a format string or callable, but got: {}".format(fmt))
            else:
                raise StructError("Internal error: Expected key, value[, format], but got: {}".format(field))
            if hasattr(me, key):
                raise StructError("Internal error: Duplicate Struct key {}".format(key))
            setattr(me, key, value)
            me.fields[key] = fmt
        return me

    @classmethod
    def unpack(cls, u):
        return cls(cls._unpack(u))

    def format_field(self, name):
        return self.fields[name](getattr(self, name))

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, ", ".join("{}={}".format(k, self.format_field(k)) for k in self.fields.iterkeys()))

    def __iter__(self):
        return (getattr(self, k) for k in self.fields.iterkeys())

def format_each(fmt_one):
    def f(it):
        return "({})".format(", ".join(fmt_one.format(i) for i in it))
    return f

format_each_hex = format_each("{:#x}")

def format_table(fmt, table, default='Reserved'):
    def f(value):
        return "{} ({})".format(fmt.format(value), table.get(value, default))
    return f

def format_function(fmt, function):
    def f(value):
        return "{} ({})".format(fmt.format(value), function(value))
    return f

def reserved_None(fmt="{!r}"):
    def f(value):
        if value is None:
            return "Reserved"
        return fmt.format(value)
    return f

class RSDP(Struct):
    @staticmethod
    def _unpack(u):
        start = u.offset
        yield 'signature', u.unpack_one("8s")
        yield 'checksum', u.unpack_one("B")
        yield 'oemid', u.unpack_one("6s")
        revision = u.unpack_one("B")
        yield 'revision', revision, "{}"
        yield 'rsdt_address', u.unpack_one("<I")
        if revision >= 2:
            length = u.unpack_one("<I")
            yield 'length', length, "{}"
            yield 'xsdt_address', u.unpack_one("<Q")
            yield 'extended_checksum', u.unpack_one("B")
            if revision == 2:
                reserved = u.unpack_raw(3)
            else:
                yield 'unknown', u.unpack_raw(length - (u.offset - start))

parse_rsdp = make_compat_parser("RSDP")

class RSDT(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        def _tables():
            while not u.at_end():
                yield u.unpack_one("<I")
        yield 'tables', tuple(_tables()), format_each_hex

parse_rsdt = make_compat_parser("RSDT")

class XSDT(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        def _tables():
            while not u.at_end():
                yield u.unpack_one("<Q")
        yield 'tables', tuple(_tables()), format_each_hex

parse_xsdt = make_compat_parser("XSDT")

ACPI_DMAR_TYPE_DRHD = 0
ACPI_DMAR_TYPE_RMRR = 1
ACPI_DMAR_TYPE_ATSR = 2
ACPI_DMAR_TYPE_RHSA = 3

class DMAR(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        yield 'width', u.unpack_one("B"), "{}"
        flags = u.unpack_one("B")
        yield 'flags', flags
        yield 'intr_remap', bool(bitfields.getbits(flags, 0)), "flags[0]={}"
        u.skip(10)
        def _remappings():
            dmar_subtables = [
                DMARSubtableDRHD,
                DMARSubtableRMRR,
                DMARSubtableATSR,
                DMARSubtableRHSA,
                DMARSubtableUnknown, # Must always come last
            ]
            while not u.at_end():
                for s in dmar_subtables:
                    temp = s.unpack(u)
                    if temp is not None:
                        yield temp
                        break
        yield 'remappings', tuple(_remappings()), format_each("\n{!r}")

parse_dmar = make_compat_parser("DMAR")

class DMARSubtable(Struct):
    @classmethod
    def unpack(cls, u):
        t, length = u.unpack_peek("<HH")
        if cls.dmar_subtable_type is not None and t != cls.dmar_subtable_type:
            return None
        return super(DMARSubtable, cls).unpack(u.unpack_unpackable(length))

class DMARSubtableDRHD(DMARSubtable):
    dmar_subtable_type = ACPI_DMAR_TYPE_DRHD

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("<H"), "{}"
        yield 'length', u.unpack_one("<H"), "{}"
        flags = u.unpack_one("B")
        yield 'flags', flags
        yield 'include_pci_all', bitfields.getbits(flags, 0)
        u.skip(1)
        yield 'segment_number', u.unpack_one("<H")
        yield 'base_address', u.unpack_one("<Q")
        def _device_scopes():
           while not u.at_end():
                yield DMARDeviceScope.unpack(u)
        yield 'device_scopes', tuple(_device_scopes()), format_each("\n{!r}")

class DMARSubtableRMRR(DMARSubtable):
    dmar_subtable_type = ACPI_DMAR_TYPE_RMRR

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("<H"), "{}"
        yield 'length', u.unpack_one("<H"), "{}"
        u.skip(2)
        yield 'segment_number', u.unpack_one("<H")
        yield 'base_address', u.unpack_one("<Q")
        yield 'limit_address', u.unpack_one("<Q")
        def _device_scopes():
            while not u.at_end():
                yield DMARDeviceScope.unpack(u)
        yield "device_scopes", tuple(_device_scopes()), format_each("\n{!r}")

class DMARSubtableATSR(DMARSubtable):
    dmar_subtable_type = ACPI_DMAR_TYPE_ATSR

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("<H"), "{}"
        yield 'length', u.unpack_one("<H"), "{}"
        flags = u.unpack_one("B")
        yield 'flags', flags
        yield 'all_ports', bitfields.getbits(flags, 0), "flags[0]={}"
        u.skip(1)
        yield "segment_number", u.unpack_one("<H")
        def _device_scopes():
            while not u.at_end():
                yield DMARDeviceScope.unpack(u)
        yield "device_scopes", tuple(_device_scopes()), format_each("\n{!r}")

class DMARSubtableRHSA(DMARSubtable):
    dmar_subtable_type = ACPI_DMAR_TYPE_RHSA

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("<H"), "{}"
        yield 'length', u.unpack_one("<H"), "{}"
        u.skip(4)
        yield 'base_address', "<Q"
        yield 'proximity_domain', "<I"

class DMARSubtableUnknown(DMARSubtable):
    dmar_subtable_type = None

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("<H"), "{}"
        yield 'length', u.unpack_one("<H"), "{}"
        yield 'data', u.unpack_rest(), lambda data: "\n" + bits.dumpmem(data)

class DMARDeviceScope(Struct):
    @classmethod
    def unpack(cls, u):
        length = u.unpack_peek_one("<xB")
        return super(DMARDeviceScope, cls).unpack(u.unpack_unpackable(length))

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        u.skip(2)
        yield 'enumeration_id', u.unpack_one("B")
        yield 'start_bus_number', u.unpack_one("B")
        def _paths():
            while not u.at_end():
                yield DMARDeviceScopePath.unpack(u)
        yield 'paths', tuple(_paths()), format_each("\n{!r}")

class DMARDeviceScopePath(Struct):
    @staticmethod
    def _unpack(u):
        yield 'pci_device', u.unpack_one("B")
        yield 'pci_function', u.unpack_one("B")

class GenericRegister(Struct):
    @staticmethod
    def _unpack(u):
        yield 'DescriptorType', u.unpack_one("B"), "{}"
        yield 'ResourceLength', u.unpack_one("<H"), "{}"
        yield 'AddressSpaceId', u.unpack_one("B")
        yield 'BitWidth', u.unpack_one("B")
        yield 'BitOffset', u.unpack_one("B")
        yield 'AccessSize', u.unpack_one("B")
        yield 'Address', u.unpack_one("<Q")
        yield 'EndTagByte', u.unpack_one("B")
        yield 'Checksum', u.unpack_one("B")

FixedFuncHwReg = namedtuple('FixedFunctionalHardwareRegister', ['Type', 'VendorCode',
                            'ClassCode', 'Arg1', 'Arg0'])

class FACS(Struct):
    @staticmethod
    def _unpack(u):
        version = u.unpack_peek_one("<32xB")
        yield 'signature', u.unpack_one("4s")
        yield 'length', u.unpack_one("<I"), '{}'
        yield 'hardware_signature', u.unpack_one("<I")
        yield 'firmware_waking_vector', u.unpack_one("<I")
        global_lock = u.unpack_one("<I")
        yield 'global_lock', global_lock
        yield 'pending', bool(bitfields.getbits(global_lock, 0)), "global_lock[0]={}"
        yield 'owned', bool(bitfields.getbits(global_lock, 1)), "global_lock[1]={}"
        flags = u.unpack_one("<I")
        yield 'flags', flags
        yield 's4bios_f', bool(bitfields.getbits(flags, 0)), "flags[0]={}"
        if version >= 2:
            yield '64bit_wake_supported_f', bool(bitfields.getbits(flags, 1)), "flags[1]={}"
        if version >= 1:
            yield 'x_firmware_waking_vector', u.unpack_one("<Q")
            yield 'version', u.unpack_one("B")
        if version >= 2:
            u.skip(3)
            ospm_flags = u.unpack_one("<I")
            yield 'ospm_flags', ospm_flags
            yield '64bit_wake_f', bool(bitfields.getbits(ospm_flags, 0)), "ospm_flags[0]={}"
        # Eat the remaining bytes for completeness
        if version == 0:
            u.skip(40)
        elif version >= 1:
            u.skip(31)
        elif version >= 2:
            u.skip(24)

parse_facs = make_compat_parser("FACS")

class MCFG(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        u.skip(8)
        def _resources():
            while not u.at_end():
                yield MCFGResource.unpack(u)
        yield 'resources', tuple(_resources()), format_each("\n{!r}")

parse_mcfg = make_compat_parser("MCFG")

class MCFGResource(Struct):
    @staticmethod
    def _unpack(u):
        yield 'address', u.unpack_one("<Q")
        yield 'segment', u.unpack_one("<H"),
        yield 'start_bus', u.unpack_one("B")
        yield 'end_bus', u.unpack_one("B")
        u.skip(4)

MADT_TYPE_LOCAL_APIC = 0
MADT_TYPE_IO_APIC = 1
MADT_TYPE_INT_SRC_OVERRIDE = 2
MADT_TYPE_NMI_INT_SRC = 3
MADT_TYPE_LOCAL_APIC_NMI = 4
MADT_TYPE_LOCAL_X2APIC = 9
MADT_TYPE_LOCAL_X2APIC_NMI = 0xA
MADT_TYPE_LOCAL_GIC = 0xB
MADT_TYPE_LOCAL_GIC_DISTRIBUTOR = 0xC

def _apic_subtables(u):
    apic_subtables = [
        APICSubtableLocalAPIC,
        APICSubtableIOApic,
        APICSubtableIntSrcOverride,
        APICSubtableNmiIntSrc,
        APICSubtableLocalApicNmi,
        APICSubtableLocalx2Apic,
        APICSubtableLocalx2ApicNmi,
        APICSubtableUnknown, # Must always come last
    ]
    while not u.at_end():
        for s in apic_subtables:
            temp = s.unpack(u)
            if temp is not None:
                yield temp
                break

class APIC(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        yield 'local_apic_address', u.unpack_one("<I")
        flags = u.unpack_one("<I")
        yield 'flags', flags
        yield 'pcat_compat', bitfields.getbits(flags, 0), "flags[0]={}"
        subtables = tuple(_apic_subtables(u))
        procid_apicid = {}
        uid_x2apicid = {}
        for s in subtables:
            # accumulate the dictionaries
            if (s.type == MADT_TYPE_LOCAL_APIC) and s.enabled:
                procid_apicid[s.proc_id] = s.apic_id
            if (s.type == MADT_TYPE_LOCAL_X2APIC) and s.enabled:
                uid_x2apicid[s.uid] = s.x2apicid
        yield 'procid_apicid', procid_apicid
        yield 'uid_x2apicid', uid_x2apicid
        yield 'subtables', subtables, format_each("\n{!r}")

class _MAT(Struct):
    @staticmethod
    def _unpack(u):
        yield 'subtables', tuple(_apic_subtables(u))

class APICSubtable(Struct):
    @classmethod
    def unpack(cls, u):
        t, length = u.unpack_peek("<BB")
        if cls.apic_subtable_type is not None and t != cls.apic_subtable_type:
            return None
        return super(APICSubtable, cls).unpack(u.unpack_unpackable(length))

class APICSubtableLocalAPIC(APICSubtable):
    apic_subtable_type = MADT_TYPE_LOCAL_APIC

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        yield 'proc_id', u.unpack_one("B")
        yield 'apic_id', u.unpack_one("B")
        flags = u.unpack_one("<I")
        yield 'flags', flags
        yield 'enabled', bool(bitfields.getbits(flags, 0)), "flags[0]={}"

class APICSubtableIOApic(APICSubtable):
    apic_subtable_type = MADT_TYPE_IO_APIC

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        yield 'io_apic_id', u.unpack_one("B")
        reserved = u.unpack_raw(1)
        yield 'io_apic_addr', u.unpack_one("<I")
        yield 'global_sys_int_base', u.unpack_one("<I")

mps_inti_polarity = {
    0b00: 'Conforms to bus specifications',
    0b01: 'Active high',
    0b11: 'Active low',
}

mps_inti_trigger_mode = {
    0b00: 'Conforms to bus specifications',
    0b01: 'Edge-triggered',
    0b11: 'Level-triggered',
}

class APICSubtableIntSrcOverride(APICSubtable):
    apic_subtable_type = MADT_TYPE_INT_SRC_OVERRIDE

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        yield 'bus', u.unpack_one("B")
        yield 'source', u.unpack_one("B")
        yield 'global_sys_interrupt', u.unpack_one("<I")
        flags = u.unpack_one("<H")
        yield 'flags', flags
        yield 'polarity', bitfields.getbits(flags, 1, 0), format_table("flags[1:0]={}", mps_inti_polarity)
        yield 'trigger_mode', bitfields.getbits(flags, 3, 2), format_table("flags[3:2]={}", mps_inti_trigger_mode)

class APICSubtableNmiIntSrc(APICSubtable):
    apic_subtable_type = MADT_TYPE_NMI_INT_SRC

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        flags = u.unpack_one("<H")
        yield 'flags', flags
        yield 'polarity', bitfields.getbits(flags, 1, 0), format_table("flags[1:0]={}", mps_inti_polarity)
        yield 'trigger_mode', bitfields.getbits(flags, 3, 2), format_table("flags[3:2]={}", mps_inti_trigger_mode)
        yield 'global_sys_interrupt', u.unpack_one("<I")

class APICSubtableLocalApicNmi(APICSubtable):
    apic_subtable_type = MADT_TYPE_LOCAL_APIC_NMI

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        yield 'proc_id', u.unpack_one("B")
        flags = u.unpack_one("<H")
        yield 'flags', flags
        yield 'polarity', bitfields.getbits(flags, 1, 0), format_table("flags[1:0]={}", mps_inti_polarity)
        yield 'trigger_mode', bitfields.getbits(flags, 3, 2), format_table("flags[3:2]={}", mps_inti_trigger_mode)
        yield 'lint_num', u.unpack_one("B")

class APICSubtableLocalx2Apic(APICSubtable):
    apic_subtable_type = MADT_TYPE_LOCAL_X2APIC

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        u.skip(2)
        yield 'x2apicid', u.unpack_one("<I")
        flags = u.unpack_one("<I")
        yield 'flags', flags
        yield 'enabled', bool(bitfields.getbits(flags, 0)), "flags[0]={}"
        yield 'uid', u.unpack("<I")

class APICSubtableLocalx2ApicNmi(APICSubtable):
    apic_subtable_type = MADT_TYPE_LOCAL_X2APIC_NMI

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        flags = u.unpack_one("<H")
        yield 'flags', flags
        yield 'polarity', bitfields.getbits(flags, 1, 0), format_table("flags[1:0]={}", mps_inti_polarity)
        yield 'trigger_mode', bitfields.getbits(flags, 3, 2), format_table("flags[3:2]={}", mps_inti_trigger_mode)
        yield 'uid', u.unpack_one("<I")
        yield 'lint_num', u.unpack_one("B")
        u.skip(3)

class APICSubtableLocalGIC(APICSubtable):
    apic_subtable_type = MADT_TYPE_LOCAL_GIC

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        u.skip(2)
        yield 'gic_id', u.unpack_one("<I")
        yield 'uid', u.unpack_one("<I")
        flags = u.unpack_one("<I")
        yield 'flags', flags
        yield 'enabled', bool(bitfields.getbits(flags, 0)), "flags[0]={}"
        _performance_interrupt_mode = {
            0: 'Level-triggered',
            1: 'Edge-triggered',
        }
        yield 'performance_interrupt_mode', bitfields.getbits(flags, 1), format_table("{}", _performance_interrupt_mode)
        yield 'parking_protocol_version', u.unpack_one("<I"), {}
        yield 'performance_interrupt_gsiv', u.unpack_one("<I")
        yield 'parked_address', u.unpack_one("<Q")
        yield 'physical_base_adddress', u.unpack_one("<Q")

class APICSubtableLocalGIC(APICSubtable):
    apic_subtable_type = MADT_TYPE_LOCAL_GIC_DISTRIBUTOR

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        u.skip(2)
        yield 'gic_id', u.unpack_one("<I")
        yield 'physical_base_adddress', u.unpack_one("<Q")
        yield 'system_vector_base', u.unpack_one("<I")
        u.skip(4)

class APICSubtableUnknown(APICSubtable):
    apic_subtable_type = None

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        yield 'data', u.unpack_rest(), lambda data: "\n" + bits.dumpmem(data)

def parse_apic(printflag=False, EnabledOnly=False, instance=1):
    """Parse and optionally print an ACPI MADT table."""

    apic = parse_table("APIC", instance)
    if apic is None:
        return None, None
    if printflag:
        ttypager.ttypager(_wrap(repr(apic)))
    if EnabledOnly:
        for subtable in apic.subtables:
            if subtable.type in (MADT_TYPE_LOCAL_APIC, MADT_TYPE_LOCAL_X2APIC) and subtable.enabled:
                print_wrap(repr(subtable))
    return (apic.procid_apicid, apic.uid_x2apicid)

PMTT_SOCKET = 0
PMTT_MEMORY_CONTROLLER = 1
PMTT_DIMM = 2

class PMTT(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        u.skip(4)
        def _subtables():
            pmtt_subtables = [
                PMTTSubtableSocket,
                PMTTSubtableMemController,
                PMTTSubtableDIMM,
                PMTTSubtableUnknown, # Must always come last
            ]
            while not u.at_end():
                for s in pmtt_subtables:
                    temp = s.unpack(u)
                    if temp is not None:
                        yield temp
                        break
        yield 'subtables', tuple(_subtables()), format_each("\n{!r}")

parse_pmtt = make_compat_parser("PMTT")

class PMTTSubtable(Struct):
    @classmethod
    def unpack(cls, u):
        t, length = u.unpack_peek("<BxH")
        if cls.pmtt_subtable_type is not None and t != cls.pmtt_subtable_type:
            return None
        return super(PMTTSubtable, cls).unpack(u.unpack_unpackable(length))

class PMTTSubtableCommon(Struct):
    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        u.skip(1)
        yield 'length', u.unpack_one("<H"), "{}"
        flags = u.unpack_one("<H")
        yield 'flags', flags
        yield 'top_level_aggregator_device', bool(bitfields.getbits(flags, 0)), "flags[0]={}"
        yield 'physical_topology_element', bool(bitfields.getbits(flags, 1)), "flags[1]={}"
        _component_memory_type = {
            0b00:   'Volatile memory',
            0b01:   'Both volatile and non-volatile memory',
            0b10:   'Non-volatile memory',
        }
        yield 'component_memory_type', bitfields.getbits(flags, 3, 2), format_table("flags[3:2]={}", _component_memory_type)
        u.skip(2)

class PMTTSubtableSocket(PMTTSubtable):
    pmtt_subtable_type = PMTT_SOCKET

    @staticmethod
    def _unpack(u):
        yield 'subtable_common', PMTTSubtableCommon.unpack(u)
        yield 'socket_identifier', u.unpack_one("<H")
        u.skip(2)
        def _subtable():
            while not u.at_end():
                yield PMTTMemoryController.unpack(u)
        yield 'resources', tuple(_resources()), format_each("\n{!r}")

class PMTTSubtableMemController(PMTTSubtable):
    pmtt_subtable_type = PMTT_MEMORY_CONTROLLER

    @staticmethod
    def _unpack(u):
        yield 'subtable_common', PMTTSubtableCommon.unpack(u)
        u.skip(2)
        yield 'read_latency', u.unpack_one("<I")
        yield 'write_latency', u.unpack_one("<I")
        yield 'read_bandwidth', u.unpack_one("<I")
        yield 'write_bandwidth', u.unpack_one("<I")
        yield 'optimal_access_unit', u.unpack_one("<H")
        yield 'optimal_access_aligment', u.unpack_one("<H")
        u.skip(2)
        number_proximity_domains = u.unpack("<H")
        yield 'number_proximity_domains', number_proximity_domains
        yield 'domains', tuple(u.unpack_one("<I") for i in range(number_proximity_domains))
        def _subtable():
            while not u.at_end():
                yield PMTTPhysicalComponent.unpack(u)
        yield 'subtables', tuple(_subtables()), format_each("\n{!r}")

class PMTTSubtableDIMM(PMTTSubtable):
    pmtt_subtable_type = PMTT_DIMM

    @staticmethod
    def _unpack(u):
        yield 'subtable_common', PMTTSubtableCommon.unpack(u)
        u.skip(2)
        yield 'physical_component_id', u.unpack_one("<H")
        uskip(2)
        yield 'dimm_size', u.upack_one("<I")
        yield 'smbios_handle' ,u.unpack_one("<I")

class PMTTSubtableUnknown(PMTTSubtable):
    pmtt_subtable_type = None

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        u.skip(1)
        yield 'length', u.unpack_one("<H"), "{}"
        yield 'data', u.unpack_rest(), lambda data: "\n" + bits.dumpmem(data)

class MPST(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        yield 'pcc_id'. u.unpack_one("B")
        u.skip(3)
        memory_power_node_count = u.unpack_one("<H")
        yield 'memory_power_node_count', memory_power_node_count
        u.skip(2)
        yield 'memory_power_nodes', tuple(MPSTMemPowerNode.unpack(u) for i in range(memory_power_node_count)), format_each("\n{!r}")
        characteristics_count = y.unpack("<H")
        yield 'characteristics_count', characteristics_count, "{}"
        u.skip(2)
        yield 'characteristics', tuple(MPSTCharacteristics.unpack(u) for i in range(characteristics_count)), format_each("\n{!r}")

parse_mpst = make_compat_parser("MPST")

class MPSTMemPowerNode(Struct):
    @staticmethod
    def _unpack(cls, u):
        flags = u.unpack_one("B")
        yield 'flags', flags
        yield 'enabled', bool(bitfields.getbits(flags, 0)), "flags[0]={}"
        yield 'power_managed', bool(bitfields.getbits(flags, 1)), "flags[1]={}"
        yield 'hot_pluggable', bool(bitfields.getbits(flags, 2)), "flags[2]={}"
        u.skip(1)
        yield 'node_id', u.unpack_one("<H")
        yield 'length', u.unpack_one("<I")
        yield 'base_address_low', u.unpack_one("<I")
        yield 'base_address_high', u.unpack_one("<I")
        yield 'length_low', u.unpack_one("<I")
        yield 'length_high', u.unpack_one("<I")
        num_power_states = u.unpack_one("<I")
        yield 'num_power_states', num_power_states
        num_physical_components = u.unpack_one("<I")
        yield 'num_physical_components', num_physical_components
        yield 'memory_power_nodes', tuple(MPSTState.unpack(u) for i in range(num_power_states)), format_each("\n{!r}")
        yield 'physical_component_ids', tuple(u.unpack("<H") for i in range(num_physical_components))

class MPSTState(Struct):
    @staticmethod
    def _unpack(cls, u):
        yield 'value', u.unpack("B")
        yield 'information_index', u.unpack("B")

class MPSTCharacteristics(Struct):
    @staticmethod
    def _unpack(u):
        pss_id = u.unpack_one("B")
        yield 'pss_id', pss_id
        yield 'pss_id_value', bitfields.getbits(pss_id, 5, 0), "pss_id[5:0]={}"
        yield 'pss_id_revision', bitfields.getbits(pss_id, 7, 6), "pss_id[7:6]={}"
        flags = u.unpack_one("B")
        yield 'flags', flags
        yield 'memory_content_preserved', bool(bitfields.getbits(flags, 0)), "flags[0]={}"
        yield 'autonomous_power_state_entry', bool(bitfields.getbits(flags, 1)), "flags[1]={}"
        yield 'autonomous_power_state_exit', bool(bitfields.getbits(flags, 2)), "flags[2]={}"
        u.skip(2)
        yield 'average_power', u.unpack_one("<I")
        yield 'relative_power', u.unpack("<I")
        yield 'exit_latency', u.unpack_one("<Q")
        u.skip(8)

class MSCT(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        yield 'proximity_domain_info_offset', u.unpack_one("<I")
        yield 'max_proximity_domains', u.unpack_one("<I")
        yield 'max_clock_domains', u.unpack_one("<I")
        yield 'max_physical_address', u.unpack_one("<Q")
        def _subtables():
            while not u.at_end():
                yield MSCTProximityDomainInfo.unpack(u)
        yield 'subtables', tuple(_subtables()), format_each("\n{!r}")

parse_msct = make_compat_parser("MSCT")

class MSCTProximityDomainInfo(Struct):
    @staticmethod
    def _unpack(u):
        yield 'revision', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        yield 'proximity_domain_range_low', u.unpack_one("<I")
        yield 'proximity_domain_range_high', u.unpack_one("<I")
        yield 'max_processor_capacity', u.unpack_one("<I"), "{}"
        yield 'max_memory_capacity', u.unpack_one("<Q")

class SLIT(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        number_system_localities = u.unpack_one("<Q")
        yield 'number_system_localities', number_system_localities
        def _relative_distances():
            while not u.at_end():
                yield u.unpack_one("B")
        yield 'relative_distances', tuple(tuple(u.unpack_one("B") for j in range(number_system_localities)) for i in range(number_system_localities))

parse_slit = make_compat_parser("SLIT")

SRAT_LOCAL_APIC_AFFINITY = 0
SRAT_MEMORY_AFFINITY = 1
SRAT_LOCAL_X2APIC_AFFINITY = 2

class SRAT(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        u.skip(4)
        u.skip(8)
        def _subtables():
            srat_subtables = [
                SRATLocalApicAffinity,
                SRATMemoryAffinity,
                SRATLocalX2ApicAffinity,
                SRATSubtableUnknown, # Must always come last
            ]
            while not u.at_end():
                for s in srat_subtables:
                    temp = s.unpack(u)
                    if temp is not None:
                        yield temp
                        break
        yield 'subtables', tuple(_subtables()), format_each("\n{!r}")

class SRATSubtable(Struct):
    @classmethod
    def unpack(cls, u):
        t, length = u.unpack_peek("<BB")
        if cls.srat_subtable_type is not None and t != cls.srat_subtable_type:
            return None
        return super(SRATSubtable, cls).unpack(u.unpack_unpackable(length))

class SRATLocalApicAffinity(SRATSubtable):
    srat_subtable_type = SRAT_LOCAL_APIC_AFFINITY

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B")
        yield 'length', u.unpack_one("B")
        proximity_domain_7_0 = u.unpack_one("B")
        yield 'proximity_domain_7_0', proximity_domain_7_0
        yield 'apic_id', u.unpack_one("B")
        flags = u.unpack_one("<I")
        yield 'flags', flags
        yield 'enabled', bool(bitfields.getbits(flags, 0)), "flags[0]={}"
        yield 'local_sapic_eid', u.unpack_one("B")
        pd = u.unpack("3B")
        proximity_domain_31_8 = (pd[0] << 16) + (pd[1] << 8) + pd[2]
        yield 'proximity_domain_31_8', proximity_domain_31_8
        yield 'proximity_domain', proximity_domain_31_8 << 8 + proximity_domain_7_0
        yield 'clock_domain', u.unpack_one("<I")

class SRATMemoryAffinity(SRATSubtable):
    srat_subtable_type = SRAT_MEMORY_AFFINITY

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        yield 'proximity_domain', u.unpack_one("<I")
        u.skip(2)
        yield 'base_address_low', u.unpack_one("<I")
        yield 'base_address_high', u.unpack_one("<I")
        yield 'length_low', u.unpack_one("<I")
        yield 'length_high', u.unpack_one("<I")
        u.skip(4)
        flags = u.unpack_one("<I")
        yield 'flags', flags
        yield 'enabled', bool(bitfields.getbits(flags, 0)), "flags[0]={}"
        yield 'hot_pluggable', bool(bitfields.getbits(flags, 1)), "flags[1]={}"
        yield 'nonvolatile', bool(bitfields.getbits(flags, 2)), "flags[2]={}"
        u.skip(8)

class SRATLocalX2ApicAffinity(SRATSubtable):
    srat_subtable_type = SRAT_LOCAL_X2APIC_AFFINITY

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        u.skip(2)
        yield 'proximity_domain', u.unpack_one("<I")
        yield 'x2apic_id', u.unpack_one("<I")
        yield 'flags', u.unpack_one("<I")
        yield 'enabled', bool(bitfields.getbits(flags, 0)), "flags[0]={}"
        yield 'clock_domain', u.unpack_one("<I")
        u.skip(4)

class SRATSubtableUnknown(SRATSubtable):
    srat_subtable_type = None

    @staticmethod
    def _unpack(u):
        yield 'type', u.unpack_one("B"), "{}"
        yield 'length', u.unpack_one("B"), "{}"
        yield 'data', u.unpack_rest(), lambda data: "\n" + bits.dumpmem(data)

def parse_srat(printflag=False, EnabledOnly=False, instance=1):
    """Parse and optionally print an SRAT table."""

    srat = parse_table("SRAT", instance)
    if srat is None:
        return None
    if printflag:
        ttypager.ttypager(_wrap(repr(srat)))
    if EnabledOnly:
        for subtable in srat.subtables:
            if subtable.type in (SRAT_LOCAL_APIC_AFFINITY, SRAT_LOCAL_X2APIC_AFFINITY) and subtable.enabled:
                print_wrap(repr(subtable))
    return srat

ASID_SYSTEM_MEMORY = 0
ASID_SYSTEM_IO = 1
ASID_PCI_CFG_SPACE = 2
ASID_EMBEDDED_CONTROLLER = 3
ASID_SMBUS = 4
ASID_PCC = 0xA
ASID_FFH = 0x7F

def _asid_str(asid):
    if asid >= 0xC0 and asid <= 0xff:
        return 'OEM Defined'
    _asid = {
        ASID_SYSTEM_MEMORY: 'System Memory',
        ASID_SYSTEM_IO: 'System IO',
        ASID_PCI_CFG_SPACE: 'PCI Configuration Space',
        ASID_EMBEDDED_CONTROLLER: 'Embedded Controller',
        ASID_SMBUS: 'SMBus',
        ASID_PCC: 'Platform Communications Channel (PCC)',
        ASID_FFH: 'Functional Fixed Hardware',
        }
    return _asid.get(asid, 'Reserved')

class GAS(Struct):
    @staticmethod
    def _unpack(u):
        yield 'address_space_id', u.unpack_one("B"), format_function("{:#x}", _asid_str)
        yield 'register_bit_width', u.unpack_one("B")
        yield 'register_bit_offset', u.unpack_one("B")
        _access_sizes = {
            0: 'Undefined',
            1: 'Byte access',
            2: 'Word access',
            3: 'Dword access',
            4: 'Qword access',
        }
        yield 'access_size', u.unpack_one("B"), format_table("{}", _access_sizes)
        yield 'address', u.unpack_one("<Q")

class SPCR(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        yield 'interface_type', u.unpack_one("B"), "{}"
        u.skip(3)
        yield 'base_address', GAS.unpack(u)
        yield 'int_type', u.unpack_one("B")
        yield 'irq', u.unpack_one("B")
        yield 'global_sys_int', u.unpack_one("<I")
        baud_rate = u.unpack_one("B")
        yield 'baud_rate', baud_rate, "{}"
        # Decode for baud rate the BIOS used for redirection
        _baud = {
            3: 9600,
            4: 19200,
            6: 57600,
            7: 115200,
        }
        yield 'baud_rate_decode', _baud.get(baud_rate), reserved_None()
        yield 'parity', u.unpack_one("B"), format_table("{}", { 0: 'No Parity' })
        yield 'stop_bits', u.unpack_one("B"), format_table("{}", { 1: '1 stop bit' })
        flow_control = u.unpack_one("B")
        yield 'flow_control', flow_control, "{}"
        yield 'DCD', bool(bitfields.getbits(flow_control, 0))
        yield 'RTSCTS', bool(bitfields.getbits(flow_control, 1))
        yield 'XONXOFF', bool(bitfields.getbits(flow_control, 2))
        yield 'terminal_type', u.unpack_one("B"), "{}"
        u.skip(1)
        yield 'pci_did', u.unpack_one("<H")
        yield 'pci_vid', u.unpack_one("<H")
        yield 'pci_bus', u.unpack_one("B")
        yield 'pci_dev', u.unpack_one("B")
        yield 'pci_func', u.unpack_one("B")
        yield 'pci_flags', u.unpack_one("<I")
        yield 'pci_segment', u.unpack_one("B")
        u.skip(4)

parse_spcr = make_compat_parser("SPCR")

class HPET(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        event_timer_block_id = u.unpack_one("<I")
        yield 'event_timer_block_id', event_timer_block_id
        yield 'pci_vid', bitfields.getbits(event_timer_block_id, 31, 16), "event_timer_block_id[31:16]={:#x}"
        yield 'legacy_replacement_IRQ_routing_capable', bool(bitfields.getbits(event_timer_block_id, 15)), "event_timer_block_id[15]={}"
        yield 'count_size_cap_counter_size', bitfields.getbits(event_timer_block_id, 13), "event_timer_block_id[13]={}"
        yield 'num_comparators', bitfields.getbits(event_timer_block_id, 12, 8), "event_timer_block_id[12:8]={:#x}"
        yield 'hardware_rev_id', bitfields.getbits(event_timer_block_id, 7, 0), "event_timer_block_id[7:0]={:#x}"
        yield 'base_address', GAS.unpack(u)
        yield 'hpet_number', u.unpack_one("B"), "{}"
        yield 'main_counter_min_clock_tick_in_periodic_mode', u.unpack_one("<H"), "{}"
        capabilities = u.unpack_one("B")
        yield 'capabilities', capabilities
        _page_protection_table = {
            0: 'No Guarantee for page protection',
            1: '4KB page protected',
            2: '64KB page protected',
        }
        yield 'page_protection', bitfields.getbits(capabilities, 3, 0), format_table("capabilities[3:0]={:#x}", _page_protection_table)
        yield 'oem_attributes', bitfields.getbits(capabilities, 7, 4), "capabilities[7:4]={:#x}"

parse_hpet = make_compat_parser("HPET")

class WDDT(Struct):
    @staticmethod
    def _unpack(u):
        yield 'header', TableHeader.unpack(u)
        yield 'tco_spec_version', u.unpack_one("<H")
        yield 'tco_description_table_version', u.unpack_one("<H")
        yield 'pci_vid', u.unpack_one("<H")
        yield 'tco_base_address', GAS.unpack(u)
        yield 'timer_min_count', u.unpack_one("<H")
        yield 'timer_max_count', u.unpack_one("<H")
        yield 'timer_count_period', u.unpack_one("<H")
        status = u.unpack_one("<H")
        yield 'status', status
        _wdt_available_decode = {
            0: 'permanently disabled',
            1: 'available',
        }
        yield 'wdt_available', bitfields.getbits(status,0), format_table("{}", _wdt_available_decode)
        _wdt_active_decode = {
            0: 'WDT stopped when BIOS hands off control',
            1: 'WDT running when BIOS hnads off control',
        }
        yield 'wdt_active', bitfields.getbits(status,1), format_table("{}", _wdt_active_decode)
        _ownership_decode = {
            0: 'TCO is owned by the BIOS',
            1: 'TCO is owned by the OS',
        }
        yield 'ownership', bitfields.getbits(status,2), format_table("{}", _ownership_decode)
        yield 'user_reset_event', bool(bitfields.getbits(status,11))
        yield 'wdt_event', bool(bitfields.getbits(status,12))
        yield 'power_fail_event', bool(bitfields.getbits(status,13))
        yield 'unknown_reset_event', bool(bitfields.getbits(status,14))
        capability = u.unpack_one("<H")
        yield 'capability', capability
        yield 'auto_reset', bool(bitfields.getbits(capability,0))
        yield 'alert_support', bool(bitfields.getbits(capability,1))
        yield 'platform_directed_shutdown', bool(bitfields.getbits(capability,2))
        yield 'immediate_shutdown', bool(bitfields.getbits(capability,3))
        yield 'bios_handoff_support', bool(bitfields.getbits(capability,4))

parse_wddt = make_compat_parser("WDDT")

class TableHeader(Struct):
    @staticmethod
    def _unpack(u):
        yield 'signature', u.unpack_one("4s")
        yield 'length', u.unpack_one("<I"), "{}"
        yield 'revision', u.unpack_one("B"), "{}"
        yield 'checksum', u.unpack_one("B")
        yield 'oemid', u.unpack_one("<6s")
        yield 'oemtableid', u.unpack_one("<8s")
        yield 'oemrevision', u.unpack_one("<I")
        yield 'creatorid', u.unpack_one("<4s")
        yield 'creatorrevision', u.unpack_one("<I")

    def repack(self):
        return struct.pack("<4sIBB6s8sI4sI", *self)

def get_cpupaths(*args):
    cpupaths, devpaths = _bits._acpi_cpupaths(*args)
    procid_apicid, uid_x2apicid = parse_apic()
    if procid_apicid is None or uid_x2apicid is None:
        return []
    enabled_cpupaths = []
    for cpupath in cpupaths + devpaths:
        procdef = evaluate(cpupath)
        uid = evaluate(cpupath + "._UID")
        if (procdef is not None and procdef.ProcId in procid_apicid) or (uid is not None and uid in uid_x2apicid):
           enabled_cpupaths.append(cpupath)
    return enabled_cpupaths

def find_procid():
    cpupaths = get_cpupaths()
    cpupath_procid = {}
    for cpupath in cpupaths:
        processor = evaluate(cpupath)
        if processor is not None:
            cpupath_procid[cpupath] = processor.ProcId
        else:
            cpupath_procid[cpupath] = None
    return OrderedDict(sorted(cpupath_procid.items()))

def find_uid():
    cpupaths = get_cpupaths()
    cpupath_uid = {}
    for cpupath in cpupaths:
        value = evaluate(cpupath + "._UID")
        cpupath_uid[cpupath] = value
    return OrderedDict(sorted(cpupath_uid.items()))

def commonprefix(l):
    """Return the common prefix of a list of strings."""
    if not l:
        return ''
    prefix = l[0]
    for s in l[1:]:
        for i, c in enumerate(prefix):
            if c != s[i]:
                prefix = s[:i]
                break
    return prefix

def factor_commonprefix(l):
    if not l:
        return ''
    if len(l) == 1:
        return l[0]
    prefix = commonprefix(l)
    prefixlen = len(prefix)
    return prefix + "{" + ", ".join([s[prefixlen:] for s in l]) + "}"

def display_cpupaths():
    """Find and display ACPI processor name space paths"""
    cpupaths = get_cpupaths()
    print_wrap(factor_commonprefix(cpupaths))

def display_cpu_info():
    cpupaths = get_cpupaths()
    cpupath_procid = find_procid()
    cpupath_uid = find_uid()
    procid_apicid, uid_x2apicid = parse_apic()
    if procid_apicid is None or uid_x2apicid is None:
        return
    socketindex_cpuscope = {}
    for cpupath in cpupaths:
        s = cpupath
        def socket_str(apicid):
            socket_index = bits.socket_index(apicid)
            if socket_index is None:
                return ''
            return ', socketIndex=0x%02x' % socket_index
        def apicid_str(apicid):
            if apicid is None:
                return 'no ApicID'
            return 'ApicID=0x%02x%s' % (apicid, socket_str(apicid))
        procid = cpupath_procid.get(cpupath, None)
        if procid is not None:
            s += ' ProcID=%-2u (%s) ' % (procid, apicid_str(procid_apicid.get(procid, None)))
        uid = cpupath_uid.get(cpupath, None)
        if uid is not None:
            s += '_UID=%s (%s)' % (uid, apicid_str(uid_x2apicid.get(uid, None)))
        print_wrap(s)
        socketindex_cpuscope.setdefault(bits.socket_index(procid_apicid.get(procid, None)), []).append(scope(cpupath))
    for value, scopes in socketindex_cpuscope.iteritems():
        unique_scopes = set(scopes)
        print_wrap('socket {0} contains {1} processors and {2} ACPI scope: {3}'.format(value, len(scopes), len(unique_scopes), ','.join(sorted(unique_scopes))))

def display_acpi_method(method, print_one):
    """Helper function that performs all basic processing for evaluating an ACPI method"""
    cpupaths = get_cpupaths()
    uniques = {}
    for cpupath in cpupaths:
        value = evaluate(cpupath + "." + method)
        uniques.setdefault(value, []).append(cpupath)

    print_wrap("%u unique %s values" % (len(uniques), method))
    for value, cpupaths in sorted(uniques.iteritems(), key=(lambda (k,v): v)):
        print
        print_wrap(factor_commonprefix(cpupaths))
        if value is None:
            print "No %s found for these CPUs" % method
        else:
            print_one(value)

def display_csd():
    """Find and display _CSD (C-State Dependency) methods"""
    def print_csd(csd):
        if not isinstance(csd, tuple):
            print "No _CSD found for these CPUs"
            return
        fmt_str = "              %-10s %-10s %-8s %-10s %-9s %-10s"
        print fmt_str % ("NumEntries", "Revision", "Dom", "CoordType", "NumProcs", "Index")
        for csd_num, [NumEntries, Revision, Dom, CoordType, NumProcessors, Index] in enumerate(csd):
            fmt_str = "Dependency[%u] 0x%-8x 0x%-8x 0x%-6x 0x%-8x 0x%-7x %-8x"
            print fmt_str % (csd_num, NumEntries, Revision, Dom, CoordType, NumProcessors, Index)
    display_acpi_method("_CSD", print_csd)

def makeFFH(r):
    return FixedFuncHwReg(Type=r.AddressSpaceId, VendorCode=r.BitWidth,
                          ClassCode=r.BitOffset, Arg1=r.AccessSize,
                          Arg0=r.Address)

class CState(namedtuple('CState', ['Register', 'Type', 'Latency', 'Power'])):
    __slots__ = ()
    def __repr__(self):
        r = self.Register
        if r.AddressSpaceId == 0x7f:
            r = makeFFH(self.Register)
        return "Cstate(Register={reg}, Type={Type}, Latency={Latency}, Power={Power})".format(reg = r, **self._asdict())

def parse_cst():
    """Parse the _CST for all logical processors."""
    cpupaths = get_cpupaths()
    uniques = {}
    for cpupath in cpupaths:
        pkg = evaluate(cpupath + "._CST")
        if pkg is not None:
            cstate_count = pkg[0]
            cstates = tuple( CState(GenericRegister.unpack(Unpackable(subpkg[0])), subpkg[1], subpkg[2], subpkg[3]) for subpkg in pkg[1:] )
            if cstate_count != len(cstates):
                raise TableParseException('_CST parse error: State count = {} but table had {} entries'.format(cstate_count, len(cstates)))
        else:
            cstates = None
        uniques.setdefault(cstates, []).append(cpupath)
    return uniques

def print_cst(cstates):
    if not cstates:
        print "Empty _CST was returned."
        return
    print "    {:40s}  {:7s}  {:12s}  {:10s}".format("Description", "OS Type", "Latency (us)", "Power (mW)")
    for cstate_num, cstate_data in enumerate(cstates):
        desc = ''
        if cstate_data.Register.AddressSpaceId == 0x7f:
            ffh = makeFFH(cstate_data.Register)
            if ffh.VendorCode == 0:
                desc += "C1 Halt"
            elif (ffh.VendorCode == 1) and (ffh.ClassCode == 1):
                desc += "C1 I/O then Halt I/O port address = {#x}".format(Register.Arg0)
            elif (ffh.VendorCode == 1) and (ffh.ClassCode == 2):
                desc += "MWAIT {:#02x} ({})".format(ffh.Arg0, cpulib.mwait_hint_to_cstate(ffh.Arg0))
                desc += " {} {}BMAvoid".format(("SWCoord", "HWCoord")[bool(ffh.Arg1 & 1)], ("!", "")[bool(ffh.Arg1 & (1 << 1))])
        print "C{cstate_num:<d}  {desc:40.40s}  ACPI C{Type:<1d}  {Latency:<12d}  {Power:<10d}".format(cstate_num=cstate_num, desc=desc, **CState(*cstate_data)._asdict())

def display_cst():
    """Find and display _CST (C-States) methods"""
    def print_cst(cst):
        for (reg, cst_type, latency, power) in cst[1:]:
            r = GenericRegister.unpack(Unpackable(reg))
            print
            if r.AddressSpaceId == 0x7f:
                # Decode register as FFH
                print "Fixed Functional Hardware (FFH) detected, using Intel-specific decode"
                ffh = makeFFH(r)
                fmt_str = "%-5s %-10s %-9s %-10s %-8s"
                print fmt_str % ("Type", "VendorCode", "ClassCode", "Arg0", "Arg1")
                fmt_str = "0x%-3x 0x%-8x 0x%-7x 0x%-8x 0x%-8x"
                print fmt_str % (ffh.Type, ffh.VendorCode, ffh.ClassCode, ffh.Arg0, ffh.Arg1)
                if ffh.VendorCode == 0:
                    print "  C1 Halt"
                elif (ffh.VendorCode == 1) and (ffh.ClassCode == 1):
                    print "  C1 I/O then Halt I/O port address = 0x%x" % Arg0
                elif (ffh.VendorCode == 1) and (ffh.ClassCode == 2):
                    fmt_str = "  Native C-state instruction, Hint: 0x%x (%s)"
                    print fmt_str % (ffh.Arg0, cpulib.mwait_hint_to_cstate(ffh.Arg0))
                    fmt_str = "  HW coordination=%u, Bus Master avoidance=%u"
                    print fmt_str % (bool(ffh.Arg1 & 1), bool((ffh.Arg1 & (1 << 1))))
            else:
                # Decode register as actual hardware resource
                fmt_str = "%-11s %-10s %-9s %-10s %-8s"
                print fmt_str % ("AddrSpaceId", "BitWidth", "BitOffset", "AccessSize", "Address")
                fmt_str = "0x%-9x 0x%-8x 0x%-7x 0x%-8x 0x%-8x"
                print fmt_str % (r.AddressSpaceId, r.BitWidth, r.BitOffset, r.AccessSize, r.Address)
            # Decode and print ACPI c-state, latency, & power
            fmt_str = "%-7s %-7s %-7s"
            print fmt_str % ("C-state", "Latency", "Power")
            cst_type_str = "ACPI C%u" % cst_type
            latency_str = "%uus" % latency
            power_str = "%umW" % power
            print fmt_str % (cst_type_str, latency_str, power_str)
    display_acpi_method("_CST", print_cst)

def display_pct():
    """Find and display _PCT (Performance Control) methods"""
    def print_pct(pct):
        fmt_str = "%-12s %-5s %-10s %-9s %-10s %-8s"
        print fmt_str % ("", "Type",        "VendorCode","ClassCode", "Arg0",       "Arg1")
        reg_type = ("Control", "Status")
        for reg_num, reg in enumerate(pct):
            r = GenericRegister.unpack(Unpackable(reg))
            fmt_str = "%-12s 0x%-3x 0x%-8x 0x%-7x 0x%-8x 0x%-8x"
            print fmt_str % (reg_type[reg_num], r.AddressSpaceId, r.BitWidth, r.BitOffset, r.AccessSize, r.Address)
    display_acpi_method("_PCT", print_pct)

def display_pdl():
    """Find and display _PDL (P-State Depth Limit) methods"""
    def print_pdl(pdl):
        print "Lowest performance state that OSPM can use = %u" % pdl
    display_acpi_method("_PDL", print_pdl)

def display_ppc():
    """Find and display _PPC (Performance Present Capabilities) methods"""
    def print_ppc(ppc):
        print "Highest performance state that OSPM can use = %u" % ppc
    display_acpi_method("_PPC", print_ppc)

def display_psd():
    """Find and display _PSD (P-State Dependency) methods"""
    def print_psd(psd):
        # Verify the ACPI data returned is a package and not a NULL
        if not isinstance(psd, tuple):
            print "No _PSD found for these CPUs"
            return
        fmt_str = "               %-10s %-10s %-10s %-10s %-10s"
        print fmt_str % ("NumEntries", "Revision", "Dom", "CoordType", "NumProcessors")
        for psd_num, [NumEntries, Revision, Dom, CoordType, NumProcessors] in enumerate(psd):
            fmt_str = "Dependency[%u]: 0x%-8x 0x%-8x 0x%-8x 0x%-8x 0x%-8x"
            print fmt_str % (psd_num, NumEntries, Revision, Dom, CoordType, NumProcessors)
    display_acpi_method("_PSD", print_psd)

class PSD(namedtuple('PSD', ['NumEntries', 'Revision', 'Domain', 'CoordType', 'NumProcessors'])):
    __slots__ = ()
    def __repr__(self):
        return "_PSD(NumEntries={NumEntries}, Revision={Revision}, Domain={Domain:#x}, CoordType={CoordType:#x}, NumProcessors={NumProcessors})".format(**self._asdict())

def parse_psd():
    """Parse the _PSD for all logical processors."""
    cpupaths = get_cpupaths()
    uniques = {}
    for cpupath in cpupaths:
        value = evaluate(cpupath + "._PSD")
        if value is not None:
            value = PSD(*(value)[0])
        uniques.setdefault(value, []).append(cpupath)
    return uniques

class PState(namedtuple('PState', ['CoreFrequency', 'Power', 'Latency', 'BusMasterLatency', 'Control', 'Status'])):
    __slots__ = ()
    def __repr__(self):
        return "Pstate(CoreFrequency={CoreFrequency} MHz, Power={Power} mW, Latency={Latency} us, BusMasterLatency={BusMasterLatency} us, Control={Control:#x}, Status={Status:#x})".format(**self._asdict())

def parse_pss():
    """Parse the _PSS for all logical processors."""
    cpupaths = get_cpupaths()
    uniques = {}
    for cpupath in cpupaths:
        pkg = evaluate(cpupath + "._PSS")
        if pkg is not None:
            pstates = tuple(PState(*subpkg) for subpkg in pkg)
        else:
            pstates = None
        uniques.setdefault(pstates, []).append(cpupath)
    return uniques

def show_pss(pstates):
    if not pstates:
        return "Empty _PSS"
    lines = ["    {:10s} {:10s} {:12s} {:14s} {:10s} {:10s}".format("Freq (MHz)", "Power (mW)", "Latency (us)", "BMLatency (us)", "Control", "Status")]
    for pstate_num, pstate_data in enumerate(pstates):
        lines.append("P{pstate_num:<2d} {CoreFrequency:<10d} {Power:<10d} {Latency:<12d} {BusMasterLatency:<14d} {Control:#010x} {Status:#010x}".format(pstate_num=pstate_num, **PState(*pstate_data)._asdict()))
    return lines

def print_pss(pstates):
    print "\n".join(show_pss(pstates))

def display_pss():
    """Find and display _PSS (Performance Supported States) methods"""
    display_acpi_method("_PSS", print_pss)

def display_ptc():
    """Find and display _PTC (Processor Throttling Control) methods"""
    def print_ptc(ptc):
        fmt_str = "%-12s %-13s %-12s %-11s %-12s %s"
        print fmt_str % ("Processor",  "AddrSpaceId", "BitWidth",   "BitOffset", "AccessSize", "Address")
        print fmt_str % ("Throttling", "Type",        "VendorCode", "ClassCode", "Arg0",       "Arg1")
        reg_type = ("Control", "Status")
        for reg_num, reg in enumerate(ptc):
            r = GenericRegister.unpack(Unpackable(reg))
            fmt_str = "%-12s 0x%-11x 0x%-10x 0x%-9x 0x%-10x 0x%x"
            print fmt_str % (reg_type[reg_num], r.AddressSpaceId, r.BitWidth, r.BitOffset, r.AccessSize, r.Address)
    display_acpi_method("_PTC", print_ptc)

def display_tdl():
    """Find and display _TDL (T-State Depth Limit) methods"""
    def print_tdl(tdl):
        print "Lowest power throttling state that OSPM may use = %u" % tdl
    display_acpi_method("_TDL", print_tdl)

def display_tpc():
    """Find and display _TPC (Throttling Present Capabilities) methods"""
    def print_tpc(tpc):
        print "Highest throttling state that OSPM can use = %u" % tpc
    display_acpi_method("_TPC", print_tpc)

def display_tsd():
    """Find and display _TSD (T-State Dependency) methods"""
    def print_tsd(tsd):
        # Verify the ACPI data returned is a package and not a NULL
        if not isinstance(tsd, tuple):
            print "No _TSD found for these CPUs"
            return
        fmt_str = "               %-10s %-10s %-10s %-10s    %-10s"
        print fmt_str % ("NumEntries", "Revision", "Dom", "CoordType", "NumProcessors")
        for tsd_num, [NumEntries, Revision, Dom, CoordType, NumProcessors] in enumerate(tsd):
            fmt_str = "Dependency[%u]: 0x%-8u 0x%-8u 0x%-8x 0x%-8x    %-8u"
            print fmt_str % (tsd_num, NumEntries, Revision, Dom, CoordType, NumProcessors)
    display_acpi_method("_TSD", print_tsd)

def display_tss():
    """Find and display _TSS (Throttling Supported States) methods"""
    def print_tss(tss):
        # Verify the ACPI data returned is a package and not a NULL
        if not isinstance(tss, tuple):
            print "No _TSS found for these CPUs"
            return
        fmt_str = "    %-9s %-10s %-10s %-10s %-10s"
        print fmt_str % ("Percent", "Power", "Latency", "Control", "Status")
        for tstate_num, [percent, power, latency, control, status] in enumerate(tss):
            fmt_str = "T%-2u %6u%%   0x%08x 0x%08x 0x%08x 0x%08x"
            print fmt_str % (tstate_num, percent, power, latency, control, status)
    display_acpi_method("_TSS", print_tss)

def display_uid():
    """Find and display _UID"""
    def print_uid(uid):
        print "_UID = %s" % uid
    display_acpi_method("_UID", print_uid)

def get_table(signature, instance=1):
    """Get the requested ACPI table based on signature"""
    special_get = {
        'RSDP': _bits._acpi_get_rsdp,
        'RSD PTR': _bits._acpi_get_rsdp,
        'RSD PTR ': _bits._acpi_get_rsdp,
        'RSDT': _bits._acpi_get_rsdt,
        'XSDT': _bits._acpi_get_xsdt,
    }.get(signature)
    if special_get is not None:
        if instance == 1:
            return special_get()
        return None
    return _bits._acpi_get_table(signature, instance)

def get_table_list(printflag=False):
    """Get the list of ACPI table signatures"""
    tables = itertools.takewhile(lambda table: table is not None,
                                 (get_table_by_index(index) for index in itertools.count()))
    signatures = [table[:4] for table in tables]
    signatures.extend([s for s in ['RSDP', 'RSDT', 'XSDT'] if get_table(s)])
    signatures = sorted(set(signatures))
    if printflag:
        print_wrap('ACPI Table Signatures: ' + ', '.join(signatures))
    return signatures

def dump(name=""):
    s = ''
    for path in get_objpaths(name):
        s += _wrap('{} : {!r}'.format(path, evaluate(path))) + '\n'
    return s

def dumptable(name="", instance=1):
    """Dump hexadecimal and printable ASCII bytes for an ACPI table specified by 4CC and instance"""
    s = ''
    data = get_table(name, instance)
    if data is None:
        s += "ACPI table with signature of {} and instance of {} not found.\n".format(name, instance)
        return s
    s += bits.dumpmem(data)
    return s

def dumptables():
    """Dump hexdecimal and printable ASCII bytes for all ACPI tables"""
    s = ''
    for signature in get_table_list():
        for instance in itertools.count(1,):
            data = get_table(signature, instance)
            if data is None:
                break
            s += "ACPI Table {} instance {}\n".format(signature, instance)
            s += bits.dumpmem(data)
    return s

created_explore_acpi_tables_cfg = False

def create_explore_acpi_tables_cfg():
    global created_explore_acpi_tables_cfg
    if created_explore_acpi_tables_cfg:
        return
    cfg = ""
    for signature in get_table_list():
        for instance in itertools.count(1):
            data = get_table(signature, instance)
            if data is None:
                break
            parse_method = 'parse_{}'.format(str.lower(signature), instance)
            if parse_method in globals():
                cfg += 'menuentry "Decode {} Instance {}" {{\n'.format(signature, instance)
                cfg += '    py "import acpi ; acpi.{}(printflag=True, instance={}) ; del acpi"\n'.format(parse_method, instance)
                cfg += '}\n'
            if signature in ("APIC", "SRAT"):
                title = 'Decode {} Instance {} (enabled only)'.format(signature, instance)
                cfg += 'menuentry "{}" {{\n'.format(title)
                cfg += '    set pager=1\n'
                cfg += '    echo "{}"\n'.format(title)
                cfg += '    py "import acpi ; acpi.{}(EnabledOnly=True, instance={}) ; del acpi"\n'.format(parse_method, instance)
                cfg += '    set pager=0\n'
                cfg += '}\n'
            cfg += 'menuentry "Dump {} Instance {} raw" {{\n'.format(signature, instance)
            cfg += """    py 'import ttypager, acpi; ttypager.ttypager(acpi.dumptable("{}", {})) ; del ttypager, acpi'\n""".format(signature, instance)
            cfg += '}\n'
    bits.pyfs_add_static("explore_acpi_tables.cfg", cfg)
    created_explore_acpi_tables_cfg = True

def show_checksum(signature, instance=1):
    """Compute checksum of ACPI table"""

    data = get_table(signature, instance)
    if data is None:
        print "ACPI table with signature of {} and instance of {} not found.\n".format(signature, instance)
        return

    csum = sum(ord(c) for c in data)
    print 'Full checksum is {:#x}'.format(csum)
    print '1-byte checksum is {:#x}'.format(csum & 0xff)
